@@ -1,5 +1,5 @@
 /* Declare an access pattern hint for files.
-   Copyright (C) 2010-2013 Free Software Foundation, Inc.
+   Copyright (C) 2010-2022 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,7 +12,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #include <config.h>
 #include <stdio.h>
@@ -48,7 +48,7 @@ typedef enum {
   FADVISE_NOREUSE =    POSIX_FADV_NOREUSE,
   FADVISE_DONTNEED =   POSIX_FADV_DONTNEED,
   FADVISE_WILLNEED =   POSIX_FADV_WILLNEED,
-  FADVISE_RANDOM =     POSIX_FADV_RANDOM,
+  FADVISE_RANDOM =     POSIX_FADV_RANDOM
 } fadvice_t;
 #else
 typedef enum {
@@ -57,7 +57,7 @@ typedef enum {
   FADVISE_NOREUSE,
   FADVISE_DONTNEED,
   FADVISE_WILLNEED,
-  FADVISE_RANDOM,
+  FADVISE_RANDOM
 } fadvice_t;
 #endif
 
