@@ -1,11 +1,11 @@
 /* modechange.c -- file mode manipulation
 
-   Copyright (C) 1989-1990, 1997-1999, 2001, 2003-2006, 2009-2012 Free Software
+   Copyright (C) 1989-1990, 1997-1999, 2001, 2003-2006, 2009-2022 Free Software
    Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 3 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -14,7 +14,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* Written by David MacKenzie <djm@ai.mit.edu> */
 
@@ -220,12 +220,12 @@ mode_compile (char const *mode_string)
                   {
                     octal_mode = 8 * octal_mode + *p++ - '0';
                     if (ALLM < octal_mode)
-                      return NULL;
+                      goto invalid;
                   }
                 while ('0' <= *p && *p < '8');
 
                 if (affected || (*p && *p != ','))
-                  return NULL;
+                  goto invalid;
                 affected = mentioned = CHMOD_MODE_BITS;
                 value = octal_to_mode (octal_mode);
                 flag = MODE_ORDINARY_CHANGE;
