@@ -1,6 +1,6 @@
 /* Generate random integers.
 
-   Copyright (C) 2006, 2009-2010 Free Software Foundation, Inc.
+   Copyright (C) 2006-2022 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* Written by Paul Eggert.  */
 
@@ -106,13 +106,6 @@ static inline randint shift_left (randint x)
   return HUGE_BYTES ? 0 : x << CHAR_BIT;
 }
 
-/* Return X shifted right by CHAR_BIT bits.  */
-static inline randint
-shift_right (randint x)
-{
-  return HUGE_BYTES ? 0 : x >> CHAR_BIT;
-}
-
 
 /* Consume random data from *S to generate a random number in the range
    0 .. GENMAX.  */
@@ -125,7 +118,7 @@ randint_genmax (struct randint_source *s, randint genmax)
   randint randmax = s->randmax;
   randint choices = genmax + 1;
 
-  for (;;)
+  while (1)
     {
       if (randmax < genmax)
         {
@@ -148,9 +141,9 @@ randint_genmax (struct randint_source *s, randint genmax)
           /* Increase RANDMAX by appending random bytes to RANDNUM and
              UCHAR_MAX to RANDMAX until RANDMAX is no less than
              GENMAX.  This may lose up to CHAR_BIT bits of information
-             if shift_right (RANDINT_MAX) < GENMAX, but it is not
-             worth the programming hassle of saving these bits since
-             GENMAX is rarely that large in practice.  */
+             if (HUGE_BYTES ? 0 : RANDINT_MAX >> CHAR_BIT) < GENMAX,
+             but it is not worth the programming hassle of saving
+             these bits since GENMAX is rarely that large in practice.  */
 
           i = 0;
 
@@ -205,7 +198,7 @@ randint_genmax (struct randint_source *s, randint genmax)
 void
 randint_free (struct randint_source *s)
 {
-  memset (s, 0, sizeof *s);
+  explicit_bzero (s, sizeof *s);
   free (s);
 }
 
