@@ -1,6 +1,6 @@
 /* Root device and inode number checking.
 
-   Copyright (C) 2003-2013 Free Software Foundation, Inc.
+   Copyright (C) 2003-2022 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #ifndef ROOT_DEV_INO_H
 # define ROOT_DEV_INO_H 1
@@ -22,7 +22,7 @@
 # include "same-inode.h"
 
 struct dev_ino *
-get_root_dev_ino (struct dev_ino *root_d_i);
+get_root_dev_ino (struct dev_ino *root_d_i) _GL_ATTRIBUTE_NONNULL ();
 
 /* These macros are common to the programs that support the
    --preserve-root and --no-preserve-root options.  */
@@ -35,11 +35,11 @@ get_root_dev_ino (struct dev_ino *root_d_i);
     {									\
       if (STREQ (Dirname, "/"))						\
         error (0, 0, _("it is dangerous to operate recursively on %s"),	\
-               quote (Dirname));					\
+               quoteaf (Dirname));					\
       else								\
         error (0, 0,							\
                _("it is dangerous to operate recursively on %s (same as %s)"), \
-               quote_n (0, Dirname), quote_n (1, "/"));			\
+               quoteaf_n (0, Dirname), quoteaf_n (1, "/"));		\
       error (0, 0, _("use --no-preserve-root to override this failsafe")); \
     }									\
   while (0)
