@@ -1,5 +1,5 @@
 /* chown-core.c -- core functions for changing ownership.
-   Copyright (C) 2000-2013 Free Software Foundation, Inc.
+   Copyright (C) 2000-2022 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,7 +12,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* Extracted from chown.c/chgrp.c and librarified by Jim Meyering.  */
 
@@ -26,7 +26,6 @@
 #include "chown-core.h"
 #include "error.h"
 #include "ignore-value.h"
-#include "quote.h"
 #include "root-dev-ino.h"
 #include "xfts.h"
 
@@ -68,10 +67,32 @@ chopt_init (struct Chown_option *chopt)
 }
 
 extern void
-chopt_free (struct Chown_option *chopt ATTRIBUTE_UNUSED)
+chopt_free (struct Chown_option *chopt)
 {
-  /* Deliberately do not free chopt->user_name or ->group_name.
-     They're not always allocated.  */
+  free (chopt->user_name);
+  free (chopt->group_name);
+}
+
+/* Convert the numeric user-id, UID, to a string stored in xmalloc'd memory,
+   and return it.  Use the decimal representation of the ID.  */
+
+static char *
+uid_to_str (uid_t uid)
+{
+  char buf[INT_BUFSIZE_BOUND (intmax_t)];
+  return xstrdup (TYPE_SIGNED (uid_t) ? imaxtostr (uid, buf)
+                  : umaxtostr (uid, buf));
+}
+
+/* Convert the numeric group-id, GID, to a string stored in xmalloc'd memory,
+   and return it.  Use the decimal representation of the ID.  */
+
+static char *
+gid_to_str (gid_t gid)
+{
+  char buf[INT_BUFSIZE_BOUND (intmax_t)];
+  return xstrdup (TYPE_SIGNED (gid_t) ? imaxtostr (gid, buf)
+                  : umaxtostr (gid, buf));
 }
 
 /* Convert the numeric group-id, GID, to a string stored in xmalloc'd memory,
@@ -81,11 +102,8 @@ chopt_free (struct Chown_option *chopt ATTRIBUTE_UNUSED)
 extern char *
 gid_to_name (gid_t gid)
 {
-  char buf[INT_BUFSIZE_BOUND (intmax_t)];
   struct group *grp = getgrgid (gid);
-  return xstrdup (grp ? grp->gr_name
-                  : TYPE_SIGNED (gid_t) ? imaxtostr (gid, buf)
-                  : umaxtostr (gid, buf));
+  return grp ? xstrdup (grp->gr_name) : gid_to_str (gid);
 }
 
 /* Convert the numeric user-id, UID, to a string stored in xmalloc'd memory,
@@ -95,11 +113,8 @@ gid_to_name (gid_t gid)
 extern char *
 uid_to_name (uid_t uid)
 {
-  char buf[INT_BUFSIZE_BOUND (intmax_t)];
   struct passwd *pwd = getpwuid (uid);
-  return xstrdup (pwd ? pwd->pw_name
-                  : TYPE_SIGNED (uid_t) ? imaxtostr (uid, buf)
-                  : umaxtostr (uid, buf));
+  return pwd ? xstrdup (pwd->pw_name) : uid_to_str (uid);
 }
 
 /* Allocate a string representing USER and GROUP.  */
@@ -134,18 +149,18 @@ user_group_str (char const *user, char const *group)
    CHANGED describes what (if anything) has happened. */
 
 static void
-describe_change (const char *file, enum Change_status changed,
+describe_change (char const *file, enum Change_status changed,
                  char const *old_user, char const *old_group,
                  char const *user, char const *group)
 {
-  const char *fmt;
+  char const *fmt;
   char *old_spec;
   char *spec;
 
   if (changed == CH_NOT_APPLIED)
     {
       printf (_("neither symbolic link %s nor referent has been changed\n"),
-              quote (file));
+              quoteaf (file));
       return;
     }
 
@@ -185,7 +200,7 @@ describe_change (const char *file, enum Change_status changed,
       abort ();
     }
 
-  printf (fmt, quote (file), old_spec, spec);
+  printf (fmt, quoteaf (file), old_spec, spec);
 
   free (old_spec);
   free (spec);
@@ -323,20 +338,20 @@ change_file_owner (FTS *fts, FTSENT *ent,
         }
       if (! chopt->force_silent)
         error (0, ent->fts_errno, _("cannot access %s"),
-               quote (file_full_name));
+               quoteaf (file_full_name));
       ok = false;
       break;
 
     case FTS_ERR:
       if (! chopt->force_silent)
-        error (0, ent->fts_errno, "%s", quote (file_full_name));
+        error (0, ent->fts_errno, "%s", quotef (file_full_name));
       ok = false;
       break;
 
     case FTS_DNR:
       if (! chopt->force_silent)
         error (0, ent->fts_errno, _("cannot read directory %s"),
-               quote (file_full_name));
+               quoteaf (file_full_name));
       ok = false;
       break;
 
@@ -377,7 +392,7 @@ change_file_owner (FTS *fts, FTSENT *ent,
             {
               if (! chopt->force_silent)
                 error (0, errno, _("cannot dereference %s"),
-                       quote (file_full_name));
+                       quoteaf (file_full_name));
               ok = false;
             }
 
@@ -466,7 +481,7 @@ change_file_owner (FTS *fts, FTSENT *ent,
         error (0, errno, (uid != (uid_t) -1
                           ? _("changing ownership of %s")
                           : _("changing group of %s")),
-               quote (file_full_name));
+               quoteaf (file_full_name));
     }
 
   if (chopt->verbosity != V_off)
@@ -485,11 +500,21 @@ change_file_owner (FTS *fts, FTSENT *ent,
              : CH_SUCCEEDED);
           char *old_usr = file_stats ? uid_to_name (file_stats->st_uid) : NULL;
           char *old_grp = file_stats ? gid_to_name (file_stats->st_gid) : NULL;
+          char *new_usr = chopt->user_name
+                          ? chopt->user_name : uid != -1
+                                               ? uid_to_str (uid) : NULL;
+          char *new_grp = chopt->group_name
+                          ? chopt->group_name : gid != -1
+                                               ? gid_to_str (gid) : NULL;
           describe_change (file_full_name, ch_status,
                            old_usr, old_grp,
-                           chopt->user_name, chopt->group_name);
+                           new_usr, new_grp);
           free (old_usr);
           free (old_grp);
+          if (new_usr != chopt->user_name)
+            free (new_usr);
+          if (new_grp != chopt->group_name)
+            free (new_grp);
         }
     }
 
@@ -525,7 +550,7 @@ chown_files (char **files, int bit_flags,
 
   FTS *fts = xfts_open (files, bit_flags | stat_flags, NULL);
 
-  while (1)
+  while (true)
     {
       FTSENT *ent;
 
